#  Copyright (c) 2016 Jeroen Wouters

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy as sp

# Numerical integration of ensemble of certain dynamical systems

isvector = lambda X: np.prod(X.shape)==np.max(X.shape)

class Ensemble(object): # abstract class, members should implement .evolve
	def __init__(self, x = None, M = None, t = 0.0, dt = None, measurementObservable = None, measurementDt = None):
		# MeasurementObservable should return a 2d array, where the first dimension counts the particles and the second counts the observable elements
		self.t = t
		self.dt = dt
		self.history = None
		self.measurementObservable = measurementObservable
		self.measurementDt = measurementDt
		self.measurementT = []
		
		if x is not None:
			if np.isscalar(x):
				self.x = np.array([x]) if M is None else np.array([x]*M)
			elif type(x) == list:
				self.x = np.array([x]) if M is None else np.array([x]*M)
			elif type(x) is np.ndarray:
				if isvector(x):
					self.x = x if M is None else np.array([x]*M)
				else:
					self.x = x
			else:
				self.x = []
	
	def evolve(self, t, dt = None, measurementObservable = None, measurementDt = None):
		if t < 0:
			print("Error: cannot evolve over negative time.")
			return
		if dt is None:
			dt = self.dt
		if measurementObservable is None:
			measurementObservable = self.measurementObservable
		if measurementDt is None:
			measurementDt = self.measurementDt

		if measurementDt is None or measurementDt == 0:
			numMeasurements = 0
			integrationIntervals = [t]
		else:
			numMeasurements = int(np.ceil(t/measurementDt))
			integrationIntervals = [measurementDt]*(numMeasurements-1) + [t - (numMeasurements-1)*measurementDt]
		
		if measurementObservable != None and numMeasurements > 0 and self.history is None:
			self.measurementT = [self.t]
			self.history = measurementObservable(self.x)
		
		for integrationInterval in integrationIntervals:
			self.integrate(integrationInterval, dt)
			if measurementObservable != None and numMeasurements > 0 and self.history is not None:
				self.measurementT.append(self.t)
				measurement = measurementObservable(self.x)
				self.history = np.dstack((self.history, measurement[:,np.newaxis]))
								
class Diffusion(Ensemble):
	def __init__(self, f = None, sigma = None, *args, **kwargs):
		super().__init__(*args, **kwargs)
		if np.isscalar(sigma):
			self.evolution = lambda dt: (lambda x: (x+f(x)*dt) + np.random.normal(size=x.shape)*sigma*np.sqrt(dt))
		elif isvector(sigma):
			self.evolution = lambda dt: (lambda x: (x+f(x)*dt) + self.dW(sigma)*np.sqrt(dt))
		else:
			self.evolution = lambda dt: (lambda x: (x+f(x)*dt) + np.random.normal(size=x.shape)*sigma(x)*np.sqrt(dt))
	
	def dW(self, sigma):
		mydW = np.zeros(self.x.shape)
		mydW[:,sigma != 0] = sigma[sigma != 0]*np.random.randn(self.x.shape[0], sum(sigma != 0))
		return mydW
	
	def integrate(self, t, dt = None):
		numsteps = int(np.ceil(t/dt))
		dt = t/numsteps
		evolution = self.evolution(dt)
		for step in range(numsteps):
			self.x = evolution(self.x)
		self.t = self.t + t

class Ou(Diffusion):
	"""Integrate an ensemble of Ornstein-Uhlenbeck particles."""
	
	instanton = lambda thresh, T, t: thresh*np.sinh(t)/np.sinh(T)

	# The variance of the distribution over time under OU dynamics, given initial
	# condition at 0 and initial variance 0
	nu = lambda t: (1/2.)*(1-np.exp(-2*t))
	
	def XTDecorr(self, gamma=None):
		if gamma is None:
			gamma = self.gamma
		if np.isscalar(gamma):
			return 1/gamma if gamma != 0 else 1.
		else:
			return 1/np.max(np.real(np.linalg.eigvals(gamma)))

	# Probability of over threshold value at time t is given by a rescaled erfc
	gammaThreshold = lambda thresh, t: (1/2.)*sp.special.erfc(thresh/np.sqrt(2*Ou.nu(t)))

	def __init__(self, gamma = 1., sigma = 1., *args, **kwargs):
		"""Keyword arguments:
		x	--	the initial position of the particles. If a float is given, 
				M is expected to give the number of copies, if a vector is given,
				the length of the vector determines the number of copies and M is disregarded.
		M 	-- the number of copies if x is a float and not a vector
		t 	-- the initial time
		maxDt -- the maximum time step
		measurementObservable -- a function that is applied to the particles at each time step and stored "history"
		"""
		self.gamma = gamma
		f = (lambda x: -gamma*x) if np.isscalar(gamma) else (lambda x: - np.dot(gamma,x.T).T)
		super().__init__(f = f, sigma = sigma, *args, **kwargs)
		if self.dt is None:
			self.dt = self.XTDecorr(gamma)/1e4
