#  Copyright (c) 2016 Jeroen Wouters

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import copy
import Ensemble
from Helper import TreeNode, discrete_random

# gives equidistant in y points along the instanton from 0 to thresh in time T
instantonInteractions = lambda T, N:  np.arcsinh(np.linspace(0, 1, N)*np.sinh(T))

#####################
#
# Weight functions
#
#####################

# weight and reweight functions of the exponential type
expWeights = lambda C, observable = lambda x: x: {"weight": lambda ips: np.exp(C*(observable(ips.rho.x) - observable(ips.oldx))), "reweight": lambda ips, x: np.exp(-C*observable(x))}
dummyWeights = {"weight": lambda x: x, "reweight": lambda x: x}

#####################
#
# IPS classes
#
#####################

class InteractingParticleSystem(): # child classes should implement .interact
	def __init__(self, rho = None, weight = None, reweight = None, \
				observable = None, buildTree=True, storeKilledPaths=False):
		self.buildTree = buildTree
		self.storeKilledPaths = storeKilledPaths
		if self.buildTree:
			# self.alive is a tree structure. The generation of a node in 
			# the tree corresponds to the index in the history list 
			# (i.e. the index of the selection step).
			# The data element with nodes that contain as data the index in the
			# history list of the corresponding particle
			self.tree = list(TreeNode(data=i) for i in range(len(rho.x)))
			self.alive = self.tree
			self.killed = []
		# self.history is a list of tuples, one item per selection step k,
		# with the first element of the item being the times at which 
		# measurements have been performed during the integration window
		# [t_k, t_{k+1}] and the second containing the measurements 
		self.history = []
		self.oldx = None
		
		# rho is an Ensemble, weight is a function taking two numpy arrays and returning one numpy array
		self.renormalizers = []
		self.numparticles = []
		self.numkilled = []
		self.weights = []
		self.interactionTimes = []
		
		if rho is not None:
			self.rho = rho
			
		if weight is not None:
			self.weight = weight
			
		if reweight is not None:
			self.reweight = reweight
		
		if observable is not None:
			self.observable = observable
	
	def run(self, interactions):
		self.oldx = copy.copy(self.rho.x)
		self.oldt = self.rho.t
		for deltaT in (interactions[1:] - interactions[:-1]):
			self.interactionTimes.append(self.rho.t)
			renormalizer, copies, weights = self.interact()
			self.killAndClone(copies)
			
			# logging
			self.renormalizers.append(renormalizer)
			self.weights.append(weights)
			self.numkilled.append(np.sum(copies==0))
			self.numparticles.append(len(self.rho.x))
			
			self.oldx = copy.copy(self.rho.x)
			self.oldt = self.rho.t
			self.rho.evolve(deltaT)
		self.history.append(copy.copy([self.rho.measurementT, self.rho.history]))

	def killAndClone(self, copies):
		self.rho.x = np.repeat(self.rho.x, np.int0(copies), axis=0)
		
		# Extend the genealogical tree and keep pointers to the youngest generation
		if self.buildTree:
			newalive = []
			for i in range(len(copies)):
				if copies[i] == 0 and self.storeKilledPaths:
					self.killed.append(self.alive[i])
				else:
					for j in range(copies[i]):
						survivor = TreeNode(len(newalive))
						newalive.append(survivor)
						self.alive[i].appendChild(survivor)
			self.alive = newalive
		
		self.history.append(copy.copy([self.rho.measurementT, self.rho.history]))
		self.rho.measurementT = []
		self.rho.history = None
	
	def ancestralPaths(self):
		ancestralpaths = []
		for alivenode in self.alive:
			jpaths = list(map(lambda x: x.data, alivenode.path()[1:]))
			ancestralpath = list(map(lambda j, h: np.squeeze(h[1][j]), jpaths, self.history[1:]))
			ancestralpaths.append(np.concatenate(ancestralpath))
		t = np.concatenate(list(h[0] for h in self.history))
		return t, np.array(ancestralpaths)
			
	def killedPaths(self, allGenerations=True):
		killedpaths = []
		for killednode in self.killed:
			generation = len(killednode.path())
			if allGenerations:
				killedpaths.append((self.history[generation-1][0], self.history[generation-1][1][killednode.data]))
			elif generation == len(self.history):
				killedpaths.append((self.history[generation-1][0], self.history[generation-1][1][killednode.data]))
		return killedpaths
		

class DiffusionMc(InteractingParticleSystem):
	def __init__(self, rho = None, alpha = 1, *args, **kwargs):
		# store the intial number of particles
		self.M = len(rho.x)
		self.alpha = alpha
		super().__init__(rho = rho, *args, **kwargs)
	
	def interact(self):
		x = self.rho.x
		weights = np.squeeze(self.weight(self))
		renormalizer = np.mean(weights)*np.power(len(x)/np.float(self.M), self.alpha)
		weights = weights/renormalizer
		
		# determine the number of copies of each particle after the interaction
		copies = np.int0(np.floor(weights + np.random.uniform(size = weights.shape)))
		#selection = np.repeat(range(len(x)), copies)
		
		return renormalizer, copies, weights
	
	def estimate(self, thresholds, density=False):
		# density determines whether the thresholds array should be considered
		# as bin edges for estimation of a bin probability 
		# P(x \in [thresholds[i], thresholds[i]])
		#or as lower thresholds for estimation of an over-threshold probability
		# P(x > thresholds[i])
		if not density:
			estimates = list( \
					np.sum(self.reweight( \
							ips=self, \
							x = self.oldx[self.observable(self.rho.x) >= threshold])) \
									*np.prod(self.renormalizers)/self.M \
					for threshold in thresholds)
		else:
			estimates = list( \
					np.sum(self.reweight( \
							ips=self, \
							x = self.oldx[np.logical_and((self.observable(self.rho.x) >= threshold0),\
									(self.observable(self.rho.x) < threshold1))])) \
									*np.prod(self.renormalizers)/self.M \
					for (threshold0, threshold1) in zip(thresholds[:-1], thresholds[1:]))
		return np.array(estimates)
	
	def estimateVariance(self, threshold):
		return None

